from functools import wraps
import inspect


def initializer(func):
    """
    Automatically assigns the parameters.

    >>> class process:
    ...     @initializer
    ...     def __init__(self, cmd, reachable=False, user='root'):
    ...         pass
    >>> p = process('halt', True)
    >>> p.cmd, p.reachable, p.user
    ('halt', True, 'root')
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        (
            names,
            varargs,
            varkw,
            defaults,
            kwonlyargs,
            kwonlydefaults,
            annotations,
        ) = inspect.getfullargspec(func)
        for name, arg in list(zip(names[1:], args)) + list(kwargs.items()):
            setattr(self, name, arg)

        for name, default in zip(reversed(names), reversed(defaults)):
            if not hasattr(self, name):
                setattr(self, name, default)

        func(self, *args, **kwargs)

    return wrapper



import torch
from torch import nn

from inFairness.auditor import SenSRAuditor
from inFairness.fairalgo.datainterfaces import FairModelResponse
from inFairness.utils import datautils


class SenSR(nn.Module):
    """Implementes the Sensitive Subspace Robustness (SenSR) algorithm.

    Proposed in `Training individually fair ML models with sensitive subspace robustness <https://arxiv.org/abs/1907.00020>`_

    Parameters
    ------------
        network: torch.nn.Module
            Network architecture
        distance_x: inFairness.distances.Distance
            Distance metric in the input space
        loss_fn: torch.nn.Module
            Loss function
        eps: float
            :math:`\epsilon` parameter in the SenSR algorithm
        lr_lamb: float
            :math:`\lambda` parameter in the SenSR algorithm
        lr_param: float
            :math:`\\alpha` parameter in the SenSR algorithm
        auditor_nsteps: int
            Number of update steps for the auditor to find worst-case examples
        auditor_lr: float
            Learning rate for the auditor
    """

    def __init__(
        self,
        network,
        distance_x,
        loss_fn,
        eps,
        lr_lamb,
        lr_param,
        auditor_nsteps,
        auditor_lr,
    ):

        super().__init__()

        self.distance_x = distance_x
        self.network = network
        self.loss_fn = loss_fn
        self.lambda_param = None
        self.eps = eps
        self.lr_lamb = lr_lamb
        self.lr_param = lr_param
        self.auditor_nsteps = auditor_nsteps
        self.auditor_lr = auditor_lr

        self.auditor = self.__init_auditor__()

    def __init_auditor__(self):

        auditor = SenSRAuditor(
            loss_fn=self.loss_fn,
            distance_x=self.distance_x,
            num_steps=self.auditor_nsteps,
            lr=self.auditor_lr,
        )
        return auditor

    def forward_train(self, X, Y):
        """Forward method during the training phase"""

        device = datautils.get_device(X)

        if self.lambda_param is None:
            self.lambda_param = torch.tensor(1.0, device=device)

        Y_pred = self.network(X)
        X_worst = self.auditor.generate_worst_case_examples(
            self.network, X, Y, lambda_param=self.lambda_param
        )

        self.lambda_param = torch.max(
            torch.stack(
                [
                    torch.tensor(0.0, device=device),
                    self.lambda_param
                    - self.lr_lamb * (self.eps - self.distance_x(X, X_worst).mean()),
                ]
            )
        )

        Y_pred_worst = self.network(X_worst)
        fair_loss = torch.mean(self.lr_param * self.loss_fn(Y_pred_worst, Y))

        response = FairModelResponse(loss=fair_loss, y_pred=Y_pred)
        return response

    def forward_test(self, X):
        """Forward method during the test phase"""

        Y_pred = self.network(X)
        response = FairModelResponse(y_pred=Y_pred)
        return response

    def forward(self, X, Y=None, *args, **kwargs):
        """Defines the computation performed at every call.

        Parameters
        ------------
            X: torch.Tensor
                Input data
            Y: torch.Tensor
                Expected output data

        Returns
        ----------
            output: torch.Tensor
                Model output
        """

        if self.training:
            return self.forward_train(X, Y)
        else:
            return self.forward_test(X)



import numpy as np
import torch
from sklearn.decomposition import TruncatedSVD
from typing import List

from inFairness.distances.mahalanobis_distance import MahalanobisDistances
from inFairness.utils import datautils


class SensitiveSubspaceDistance(MahalanobisDistances):
    """Implements Sensitive Subspace metric base class that accepts the
    basis vectors of a sensitive subspace, and computes a projection
    that ignores the sensitive subspace.

    The projection from the sensitive subspace basis vectors (A) is computed as:

    .. math:: P^{'} = I - (A (A A^{T})^{-1} A^{T})
    """

    def __init__(self):
        super().__init__()

    def fit(self, basis_vectors):
        """Fit Sensitive Subspace Distance metric

        Parameters
        --------------
            basis_vectors: torch.Tensor
                Basis vectors of the sensitive subspace
        """

        sigma = self.compute_projection_complement(basis_vectors)
        super().fit(sigma)

    def compute_projection_complement(self, basis_vectors):
        """Compute the projection complement of the space
        defined by the basis_vectors:

        projection complement given basis vectors (A) is computed as:

        .. math:: P^{'} = I - (A (A A^{T})^{-1} A^{T})

        Parameters
        -------------
            basis_vectors: torch.Tensor
                Basis vectors of the sensitive subspace
                Dimension (d, k) where d is the data features dimension
                and k is the number of protected dimensions

        Returns
        ----------
            projection complement: torch.Tensor
                Projection complement computed as described above.
                Shape (d, d) where d is the data feature dimension
        """

        # Computing the orthogonal projection
        # V(V V^T)^{-1} V^T
        projection = torch.linalg.inv(torch.matmul(basis_vectors.T, basis_vectors))

        projection = torch.matmul(basis_vectors, projection)

        # Shape: (n_features, n_features)
        projection = torch.matmul(projection, basis_vectors.T)

        # Complement the projection as: (I - Proj)
        projection_complement_ = torch.eye(projection.shape[0]) - projection
        projection_complement_ = projection_complement_.detach()

        return projection_complement_


class SVDSensitiveSubspaceDistance(SensitiveSubspaceDistance):
    """Sensitive Subspace metric that uses SVD to find the basis vectors of
    the sensitive subspace. The metric learns a subspace from a set of
    user-curated comparable data samples.

    Proposed in Section B.2 of Training individually fair ML models
    with sensitive subspace robustness

    References
    -------------
        `Yurochkin, Mikhail, Amanda Bower, and Yuekai Sun. "Training individually fair
        ML models with sensitive subspace robustness." arXiv preprint arXiv:1907.00020 (2019).`
    """

    def __init__(self):
        super().__init__()
        self.n_components_ = None

    def fit(self, X_train, n_components, autoinfer_device=True):
        """Fit SVD Sensitive Subspace distance metric parameters

        Parameters
        -------------
            X_train: torch.Tensor | List[torch.Tensor]
                Training data containing comparable data samples.
                If only one set of comparable data samples is provided, the input
                should be a torch.Tensor of shape :math:`(N, D)`. For multiple sets
                of comparable data samples a list of shape
                :math:`[ (N_1, D), \\cdots, (N_x, D)]` can be provided.
            n_components: int
                Desired number of latent variable dimensions
            autoinfer_device: bool
                Should the distance metric be automatically moved to an appropriate
                device (CPU / GPU) or not? If set to True, it moves the metric
                to the same device `X_train` is on. If set to False, keeps the metric
                on CPU.
        """

        self.n_components_ = n_components
        basis_vectors = self.compute_basis_vectors(X_train, n_components)
        super().fit(basis_vectors)

        if autoinfer_device:
            device = datautils.get_device(X_train)
            super().to(device)

    def __process_input_data__(self, X_train):
        """Process metric training data to convert from tensor to numpy and
        remove the mean and concatenate if multiple sets of training data
        is provided
        """

        if isinstance(X_train, torch.Tensor) or isinstance(X_train, np.ndarray):
            X_train = datautils.convert_tensor_to_numpy(X_train)
            return X_train

        if isinstance(X_train, list):
            X_train = [datautils.convert_tensor_to_numpy(X) for X in X_train]

            # Subtract mean and concatenate all sets of features
            X_norm = np.vstack([X - np.mean(X, axis=0) for X in X_train])
            return X_norm

        raise TypeError(
            "Provided data `X_train` should either be Tensor, np.ndarray or a list of these."
        )

    def compute_basis_vectors(self, X_train, n_components):
        """Compute basis vectors using SVD"""

        X_train = self.__process_input_data__(X_train)
        tSVD = TruncatedSVD(n_components=n_components)
        tSVD.fit(X_train)
        basis_vectors_ = tSVD.components_.T  # Shape: (n_features, n_components)
        basis_vectors_ = torch.Tensor(basis_vectors_)
        return basis_vectors_



from inFairness.fairalgo.sensei import SenSeI
from inFairness.fairalgo.sensr import SenSR
from inFairness.fairalgo.datainterfaces import FairModelResponse

__all__ = [symb for symb in globals() if not symb.startswith("_")]



import torch
from torch import nn

from inFairness.auditor import SenSeIAuditor
from inFairness.fairalgo.datainterfaces import FairModelResponse
from inFairness.utils import datautils


class SenSeI(nn.Module):
    """Implementes the Sensitive Set Invariane (SenSeI) algorithm.

    Proposed in `SenSeI: Sensitive Set Invariance for Enforcing Individual Fairness <https://arxiv.org/abs/2006.14168>`_

    Parameters
    ------------
        network: torch.nn.Module
            Network architecture
        distance_x: inFairness.distances.Distance
            Distance metric in the input space
        distance_y: inFairness.distances.Distance
            Distance metric in the output space
        loss_fn: torch.nn.Module
            Loss function
        rho: float
            :math:`\\rho` parameter in the SenSR algorithm
        eps: float
            :math:`\epsilon` parameter in the SenSR algorithm
        auditor_nsteps: int
            Number of update steps for the auditor to find worst-case examples
        auditor_lr: float
            Learning rate for the auditor
    """

    def __init__(
        self,
        network,
        distance_x,
        distance_y,
        loss_fn,
        rho,
        eps,
        auditor_nsteps,
        auditor_lr,
    ):

        super().__init__()

        self.distance_x = distance_x
        self.distance_y = distance_y
        self.network = network
        self.loss_fn = loss_fn
        self.lamb = None
        self.rho = rho
        self.eps = eps
        self.auditor_nsteps = auditor_nsteps
        self.auditor_lr = auditor_lr

        self.auditor = self.__init_auditor__()

    def __init_auditor__(self):

        auditor = SenSeIAuditor(
            distance_x=self.distance_x,
            distance_y=self.distance_y,
            num_steps=self.auditor_nsteps,
            lr=self.auditor_lr,
        )
        return auditor

    def forward_train(self, X, Y):
        """Forward method during the training phase"""

        device = datautils.get_device(X)
        minlambda = torch.tensor(1e-5, device=device)

        if self.lamb is None:
            self.lamb = torch.tensor(1.0, device=device)
        if type(self.eps) is float:
            self.eps = torch.tensor(self.eps, device=device)

        Y_pred = self.network(X)
        X_worst = self.auditor.generate_worst_case_examples(
            self.network, X, lambda_param=self.lamb
        )
        
        dist_x = self.distance_x(X, X_worst)
        mean_dist_x = dist_x.mean()
        lr_factor = torch.maximum(mean_dist_x, self.eps) / torch.minimum(mean_dist_x, self.eps)

        self.lamb = torch.max(
            torch.stack(
                [minlambda, self.lamb + lr_factor * (mean_dist_x - self.eps)]
            )
        )

        Y_pred_worst = self.network(X_worst)
        fair_loss = torch.mean(
            self.loss_fn(Y_pred, Y) + self.rho * self.distance_y(Y_pred, Y_pred_worst)
        )

        response = FairModelResponse(loss=fair_loss, y_pred=Y_pred)
        return response

    def forward_test(self, X):
        """Forward method during the test phase"""

        Y_pred = self.network(X)
        response = FairModelResponse(y_pred=Y_pred)
        return response

    def forward(self, X, Y=None, *args, **kwargs):
        """Defines the computation performed at every call.

        Parameters
        ------------
            X: torch.Tensor
                Input data
            Y: torch.Tensor
                Expected output data

        Returns
        ----------
            output: torch.Tensor
                Model output
        """

        if self.training:
            return self.forward_train(X, Y)
        else:
            return self.forward_test(X)



from abc import ABCMeta, abstractmethod
from torch import nn


class Distance(nn.Module, metaclass=ABCMeta):
    """
    Abstract base class for model distances
    """

    def __init__(self):
        super().__init__()

    def fit(self, **kwargs):
        """
        Fits the metric parameters for learnable metrics
        Default functionality is to do nothing. Subclass
        should overwrite this method to implement custom fit
        logic
        """
        pass

    @abstractmethod
    def forward(self, x, y):
        """
        Subclasses must override this method to compute particular distances

        Returns:
             Tensor: distance between two inputs
        """



from inFairness.distances.distance import Distance
from inFairness.distances.euclidean_dists import (
    EuclideanDistance,
    ProtectedEuclideanDistance,
)
from inFairness.distances.sensitive_subspace_dist import (
    SVDSensitiveSubspaceDistance,
    SensitiveSubspaceDistance,
)
from inFairness.distances.explore_distance import EXPLOREDistance
from inFairness.distances.logistic_sensitive_subspace import (
    LogisticRegSensitiveSubspace,
)
from inFairness.distances.mahalanobis_distance import (
    MahalanobisDistances,
    SquaredEuclideanDistance,
)

__all__ = [symb for symb in globals() if not symb.startswith("_")]



from typing import Dict
import torch
from dataclasses import dataclass


@dataclass
class PostProcessingObjectiveResponse:
    """Class to store the result from a post-processing algorithm"""

    y_solution: torch.Tensor = None
    objective: Dict = None



from typing import Iterable

import torch
import numpy as np

from itertools import product


def generate_data_pairs(n_pairs, datasamples_1, datasamples_2=None, comparator=None):
    """Utility function to generate (in)comparable data pairs given data samples. Use case includes
    creating a dataset of comparable and incomparable data for the EXPLORE distance metric which
    learns from such data samples.

    Parameters
    ------------
        n_pairs: int
                Number of pairs to construct
        datasamples_1: numpy.ndarray
                Array of data samples of shape (N_1, *)
        datasamples_2: numpy.ndarray
                (Optional) array of data samples of shape (N_2, *).
                If datasamples_2 is provided, then data pairs are constructed between
                datasamples_1 and datasamples_2.
                If datasamples_2 is not provided, then data pairs are constructed within
                datasamples_1
        comparator: function
                A lambda function that given two samples returns True if they should
                be paired, and False if not.
                If `comparator` is not defined, then random data samples are paired together.
                Example: `comparator = lambda x, y: (x == y)`

    Returns
    ----------
        idxs: numpy.ndarray
                A (n_pairs, 2) shaped array with indices of data sample pairs
    """

    if datasamples_2 is None:
        datasamples_2 = datasamples_1

    nsamples_1 = datasamples_1.shape[0]
    nsamples_2 = datasamples_2.shape[0]

    if comparator is None:
        ntotal = nsamples_1 * nsamples_2
        assert (
            n_pairs <= ntotal
        ), f"Number of desired data pairs {n_pairs} is greater than possible combinations {ntotal}"

        idxs = np.random.choice(ntotal, n_pairs, replace=False)
        idxs1, idxs2 = np.unravel_index(idxs, shape=(nsamples_1, nsamples_2))
        idxs = np.stack((idxs1, idxs2), axis=-1)
    else:
        all_idxs = [
            (idx1, idx2)
            for idx1, idx2 in product(range(nsamples_1), range(nsamples_2))
            if comparator(datasamples_1[idx1], datasamples_2[idx2])
        ]
        assert n_pairs <= len(all_idxs), (
            f"Number of desired data pairs {n_pairs} is greater than possible "
            + "combinations {len(all_idxs)}"
        )
        idx_positions = np.random.choice(len(all_idxs), n_pairs, replace=False)
        idxs = np.array([all_idxs[x] for x in idx_positions])

    return idxs


def convert_tensor_to_numpy(tensor):
    """Converts a PyTorch tensor to numpy array

    If the provided `tensor` is not a PyTorch tensor, it returns the same object back
    with no modifications

    Parameters
    -----------
        tensor: torch.Tensor
            Tensor to be converted to numpy array

    Returns
    ----------
        array_np: numpy.ndarray
            Numpy array of the provided tensor
    """

    if torch.is_tensor(tensor):
        array_np = tensor.detach().cpu().numpy()
        return array_np

    return tensor


def include_exclude_terms(
    data_terms: Iterable[str], include: Iterable[str] = (), exclude: Iterable[str] = ()
):
    """
    given a set of data terms, return a resulting set depending on specified included and excluded terms.

    Parameters
    -----------
        data_terms: string iterable
                    set of terms to be filtered
        include: string iterable
                 set of terms to be included, if not specified all data_terms are included
        exclude: string iterable
                 set of terms to be excluded from data_terms
    Returns
    ----------
        terms: list of strings
               resulting terms in alphabetical order.
    """
    terms = set(include) if len(include) > 0 else set(data_terms)
    if len(exclude) > 0:
        terms = terms.difference(set(exclude))
    terms = sorted(list(terms))
    return terms


def get_device(obj):
    """Returns a device (cpu/cuda) based on the type of the reference object

    Parameters
    -------------
        obj: torch.Tensor

    """

    device = torch.device("cpu")

    # If reference object is a tensor, use its device
    if torch.is_tensor(obj):
        device = obj.device

    # If reference object is a list, check if first element is a tensor
    # and if it is a tensor, use it's device
    if isinstance(obj, list) and torch.is_tensor(obj[0]):
        device = obj[0].device

    return device



import torch


def build_graph_from_dists(
    dists: torch.Tensor,
    scale: float = None,
    threshold: float = None,
    normalize: bool = False,
):
    """Build the adjacency matrix `W` given distances

    Parameters
    -------------
        dists: torch.Tensor
            Distance matrix between data points. Shape: (N, N)
        scale: float
            Parameter used to scale the computed distances
        threshold: float
            Parameter used to determine if two data points are connected or not.
            Distances below threshold value are connected, and beyond
            threshold value are disconnected.
        normalize: bool
            Whether to normalize the adjacency matrix or not

    Returns
    ----------
        W: torch.Tensor
            Adjancency matrix. It contains data points which are connected
            to atleast one other datapoint. Isolated datapoints, or ones which
            are not connected to any other datapoints, are not included in the
            adjancency matrix.
        idxs_in: torch.Tensor
            Indices of data points which are included in the adjacency matrix
    """

    scale = 1.0 if scale is None else scale
    threshold = 1e10 if threshold is None else threshold

    W = torch.exp(-(dists * scale)) * (torch.sqrt(dists) < threshold)
    idxs_in = torch.where(W.sum(axis=1) > 0.0)[0]

    W = W[idxs_in]
    W = W[:, idxs_in]

    if normalize:
        D_inv_sqrt = 1.0 / torch.sqrt(W.sum(axis=1))
        W = W * D_inv_sqrt * D_inv_sqrt.view(-1, 1)

    return W, idxs_in


def get_laplacian(W: torch.Tensor, normalize: bool = False):
    """Get the Laplacian of the matrix `W`

    Parameters
    -------------
        W: torch.Tensor
            Adjacency matrix of shape (N, N)
        normalize: bool
            Whether to normalize the computed laplacian or not

    Returns
    -------------
        Laplacian: torch.Tensor
            Laplacian of the adjacency matrix
    """

    D = W.sum(axis=1)
    L = torch.diag(D) - W

    if normalize:
        L = L / D.view(-1, 1)

    return L


def laplacian_solve(L: torch.Tensor, y_hat: torch.Tensor, lambda_param: float = None):
    """Solves a system of linear equations to get the post-processed output.
    The system of linear equation it solves is:
    :math:`\hat{{f}} = {(I + \lambda * L)}^{-1} \hat{y}`

    Parameters
    ------------
        L: torch.Tensor
            Laplacian matrix
        y_hat: torch.Tensor
            Model predicted output class probabilities
        lambda_param: float
            Weight for the laplacian regularizer

    Returns
    ----------
        y: torch.Tensor
            Post-processed solution according to the equation above
    """

    lambda_param = 1.0 if lambda_param is None else lambda_param
    n = L.shape[0]
    y = torch.linalg.solve(lambda_param * L + torch.eye(n), y_hat)

    return y



import torch

from dataclasses import dataclass


@dataclass
class FairModelResponse:
    """Class to store a result from the fairmodel algorithm"""

    loss: torch.Tensor = None
    y_pred: torch.Tensor = None



import numpy as np


def is_tensor_binary(data: np.ndarray):
    """Checks if the data is binary (0/1) or not. Return True if it is binary data

    Parameters
    --------------
        data: np.ndarray
            Data to validata if binary or not

    Returns
    ----------
        is_binary: bool
            True if data is binary. False if not
    """

    nonbindata = (data != 0) & (data != 1)
    has_nonbin_data = True in nonbindata
    return not has_nonbin_data



from inFairness.postprocessing.glif import GraphLaplacianIF
from inFairness.postprocessing.base_postprocessing import BasePostProcessing


__all__ = [symb for symb in globals() if not symb.startswith("_")]



import torch
import numpy as np

from inFairness.utils.postprocessing import (
    build_graph_from_dists,
    get_laplacian,
    laplacian_solve,
)
from inFairness.postprocessing.base_postprocessing import BasePostProcessing
from inFairness.postprocessing.datainterfaces import PostProcessingObjectiveResponse


class GraphLaplacianIF(BasePostProcessing):
    """Implements the Graph Laplacian Individual Fairness Post-Processing method.

    Proposed in `Post-processing for Individual Fairness <https://arxiv.org/abs/2110.13796>`_

    Parameters
    ------------
        distance_x: inFairness.distances.Distance
            Distance metric in the input space
        is_output_probas: bool
            True if the `data_Y` (model output) are probabilities implying that
            this is a classification setting, and False if the `data_Y` are
            in euclidean space implying that this is a regression setting.
    """

    def __init__(self, distance_x, is_output_probas):

        super().__init__(distance_x, is_output_probas=is_output_probas)
        self._METHOD_COORDINATE_KEY = "coordinate-descent"
        self._METHOD_EXACT_KEY = "exact"

    def __exact_pp__(self, lambda_param, scale, threshold, normalize):
        """Implements Exact version of post processing"""

        y_hat = self.__get_yhat__()

        W, idxs = build_graph_from_dists(
            self.distance_matrix, scale, threshold, normalize
        )

        data_y_idxs = y_hat[idxs]
        L = get_laplacian(W, normalize)

        if normalize:
            L = (L + L.T) / 2

        y = laplacian_solve(L, data_y_idxs, lambda_param)
        data_y_new = torch.clone(y_hat)
        data_y_new[idxs] = y

        objective = self.get_objective(
            data_y_new, lambda_param, scale, threshold, normalize, W, idxs, L
        )

        return data_y_new, objective

    def __coordinate_update__(
        self,
        yhat_batch,
        W_batch,
        y,
        batchidx,
        lambda_param,
        D_inv_batch=None,
        diag_W_batch=None,
        D_batch=None,
    ):

        W_xy = W_batch.unsqueeze(-1) * y.unsqueeze(0)

        """
        Shapes:
            W_batch: (bsz, nsamples)
            y: (nsamples, ncls -  1)
            W_xy: (bsz, nsamples, ncls-1)
            W_xy_corr: (bsz, ncls-1)
            numerator: (bsz, ncls-1)
            denominator: (bsz, 1)
        """

        if D_inv_batch is None:
            W_xy_corr = torch.diagonal(W_xy[:, batchidx], offset=0, dim1=0, dim2=1).T
            numerator = yhat_batch + lambda_param * (W_xy.sum(dim=1) - W_xy_corr)
            denominator = 1 + lambda_param * (
                W_batch.sum(dim=1, keepdim=True) - diag_W_batch.view(-1, 1)
            )
            y_new = numerator / denominator

        else:
            W_xy = W_xy * D_inv_batch.unsqueeze(-1)
            W_xy_corr = torch.diagonal(W_xy[:, batchidx], offset=0, dim1=0, dim2=1).T
            numerator = yhat_batch + (lambda_param * (W_xy.sum(dim=1) - W_xy_corr)) / 2
            denominator = (
                1
                + lambda_param
                - lambda_param * diag_W_batch.view(-1, 1) / D_batch.view(-1, 1)
            )
            y_new = numerator / denominator

        return y_new

    def __coordinate_pp__(
        self, lambda_param, scale, threshold, normalize, batchsize, epochs
    ):
        """Implements coordinate descent for large-scale data"""

        y_hat = self.__get_yhat__()
        y_copy = y_hat.clone()
        n_samples = self.datastore.n_samples

        W, idxs = build_graph_from_dists(
            self.distance_matrix, scale, threshold, normalize
        )

        data_y_idxs = y_hat[idxs]
        W_diag = torch.diag(W)

        if normalize:
            D = W.sum(dim=1)
            D_inv = 1 / D.reshape(1, -1) + 1 / D.reshape(-1, 1)

        for epoch_idx in range(epochs):
            epoch_idxs_random = np.random.permutation(n_samples)
            curridx = 0

            while curridx < n_samples:
                batchidxs = epoch_idxs_random[curridx : curridx + batchsize]

                if normalize:
                    y_copy[batchidxs] = self.__coordinate_update__(
                        data_y_idxs[batchidxs],
                        W[batchidxs],
                        y_copy,
                        batchidxs,
                        lambda_param=lambda_param,
                        D_inv_batch=D_inv[batchidxs],
                        diag_W_batch=W_diag[batchidxs],
                        D_batch=D[batchidxs],
                    )
                else:
                    y_copy[batchidxs] = self.__coordinate_update__(
                        data_y_idxs[batchidxs],
                        W[batchidxs],
                        y_copy,
                        batchidxs,
                        lambda_param=lambda_param,
                        diag_W_batch=W_diag[batchidxs],
                    )

                curridx += batchsize

        pp_sol = y_hat.clone()
        pp_sol[idxs] = y_copy

        objective = self.get_objective(
            pp_sol, lambda_param, scale, threshold, normalize, W, idxs
        )

        return pp_sol, objective

    def get_objective(
        self,
        y_solution,
        lambda_param: float,
        scale: float,
        threshold: float,
        normalize: bool = False,
        W_graph=None,
        idxs=None,
        L=None,
    ):
        """Compute the objective values for the individual fairness as follows:

        .. math:: \\widehat{\\mathbf{f}} =  \\arg \\min_{\\mathbf{f}} \\ \\|\\mathbf{f} - \\hat{\\mathbf{y}}\\|_2^2 + \\lambda \\ \\mathbf{f}^{\\top}\\mathbb{L_n} \\mathbf{f}

        Refer equation 3.1 in the paper

        Parameters
        ------------
            y_solution: torch.Tensor
                Post-processed solution values of shape (N, C)
            lambda_param: float
                Weight for the Laplacian Regularizer
            scale: float
                Parameter used to scale the computed distances.
                Refer equation 2.2 in the proposing paper.
            threshold: float
                Parameter used to construct the Graph from distances
                Distances below provided threshold are considered to be
                connected edges, while beyond the threshold are considered to
                be disconnected. Refer equation 2.2 in the proposing paper.
            normalize: bool
                Whether to normalize the computed Laplacian or not
            W_graph: torch.Tensor
                Adjacency matrix of shape (N, N)
            idxs: torch.Tensor
                Indices of data points which are included in the adjacency matrix
            L: torch.Tensor
                Laplacian of the adjacency matrix

        Returns
        ---------
            objective: PostProcessingObjectiveResponse
                post-processed solution containing two parts:
                    (a) Post-processed output probabilities of shape  (N, C)
                        where N is the number of data samples, and C is the
                        number of output classes
                    (b) Objective values. Refer equation 3.1 in the paper
                        for an explanation of the various parts

        """

        if W_graph is None or idxs is None:
            W_graph, idxs = build_graph_from_dists(
                self.distance_matrix, scale, threshold, normalize
            )
        if L is None:
            L = get_laplacian(W_graph, normalize)

        y_hat = self.__get_yhat__()
        y_dist = ((y_hat - y_solution) ** 2).sum()
        L_obj = lambda_param * (y_solution[idxs] * (L @ y_solution[idxs])).sum()
        overall_objective = y_dist + L_obj

        result = {
            "y_dist": y_dist.item(),
            "L_objective": L_obj.item(),
            "overall_objective": overall_objective.item(),
        }

        return result

    def postprocess(
        self,
        method: str,
        lambda_param: float,
        scale: float,  # 0.001
        threshold: float,  # median of all distances if None
        normalize: bool = False,
        batchsize: int = None,
        epochs: int = None,
    ):
        """Implements the Graph Laplacian Individual Fairness Post-processing algorithm

        Parameters
        -------------
            method: str
                GLIF method type. Possible values are:

                (a) `coordinate-descent` method which is more suitable for
                large-scale data and post-processes by batching data into minibatches
                (see section 3.2.2 of the paper), or

                (b) `exact` method which gets the exact solution but is not appropriate
                for large-scale data (refer equation 3.3 in the paper).
            lambda_param: float
                Weight for the Laplacian Regularizer
            scale: float
                Parameter used to scale the computed distances.
                Refer equation 2.2 in the proposing paper.
            threshold: float
                Parameter used to construct the Graph from distances
                Distances below provided threshold are considered to be
                connected edges, while beyond the threshold are considered to
                be disconnected. Refer equation 2.2 in the proposing paper.
            normalize: bool
                Whether to normalize the computed Laplacian or not
            batchsize: int
                Batch size. *Required when method=`coordinate-descent`*
            epochs: int
                Number of coordinate descent epochs.
                *Required when method=`coordinate-descent`*

        Returns
        -----------
            solution: PostProcessingObjectiveResponse
                post-processed solution containing two parts:
                    (a) Post-processed output probabilities of shape  (N, C)
                        where N is the number of data samples, and C is the
                        number of output classes
                    (b) Objective values. Refer equation 3.1 in the paper
                        for an explanation of the various parts
        """

        assert method in [
            self._METHOD_COORDINATE_KEY,
            self._METHOD_EXACT_KEY,
        ], f"`method` should be either `coordinate-descent` or `exact`. Value provided: {method}"

        if method == self._METHOD_COORDINATE_KEY:
            assert (
                batchsize is not None and epochs is not None
            ), f"batchsize and epochs parameter is required but None provided"

        if method == self._METHOD_EXACT_KEY:
            data_y_new, objective = self.__exact_pp__(
                lambda_param, scale, threshold, normalize
            )
        elif method == self._METHOD_COORDINATE_KEY:
            data_y_new, objective = self.__coordinate_pp__(
                lambda_param, scale, threshold, normalize, batchsize, epochs
            )

        if self.is_output_probas:
            pp_sol = torch.exp(data_y_new) / (
                1 + torch.exp(data_y_new).sum(axis=1, keepdim=True)
            )
            y_solution = torch.hstack((pp_sol, 1 - pp_sol.sum(axis=1, keepdim=True)))
        else:
            y_solution = data_y_new

        result = PostProcessingObjectiveResponse(
            y_solution=y_solution, objective=objective
        )
        return result



import torch

from inFairness.distances.distance import Distance


class EuclideanDistance(Distance):
    def __init__(self):
        super().__init__()

    def forward(self, x, y, itemwise_dist=True):

        if itemwise_dist:
            return torch.cdist(x.unsqueeze(1), y.unsqueeze(1)).reshape(-1, 1)
        else:
            return torch.cdist(x, y)


class ProtectedEuclideanDistance(Distance):
    def __init__(self):
        super().__init__()

        self._protected_attributes = None
        self._num_attributes = None
        self.protected_vector = None

    def to(self, device):
        """Moves distance metric to a particular device

        Parameters
        ------------
            device: torch.device
        """

        assert (
            self.protected_vector is not None
        ), "Please fit the metric before moving parameters to device"

        self.device = device
        self.protected_vector = self.protected_vector.to(self.device)

    def fit(self, protected_attributes, num_attributes):
        """Fit Protected Euclidean Distance metric

        Parameters
        ------------
            protected_attributes: Iterable[int]
                List of attribute indices considered to be protected.
                The metric would ignore these protected attributes while
                computing distance between data points.
            num_attributes: int
                Total number of attributes in the data points.
        """

        self._protected_attributes = protected_attributes
        self._num_attributes = num_attributes

        self.protected_vector = torch.ones(num_attributes)
        self.protected_vector[protected_attributes] = 0.0

    def forward(self, x, y, itemwise_dist=True):
        """
        :param x, y: a B x D matrices
        :return: B x 1 matrix with the protected distance camputed between x and y
        """
        protected_x = (x * self.protected_vector).unsqueeze(1)
        protected_y = (y * self.protected_vector).unsqueeze(1)

        if itemwise_dist:
            return torch.cdist(protected_x, protected_y).reshape(-1, 1)
        else:
            return torch.cdist(protected_x, protected_y)



import torch
from torch.nn import Parameter

from inFairness.auditor.auditor import Auditor
from inFairness.utils.params import freeze_network, unfreeze_network
from inFairness.utils.datautils import get_device


class SenSeIAuditor(Auditor):
    """SenSeI Auditor implements the functionality to generate worst-case examples
    by solving the following optimization equation:

    .. math:: x_{t_b}' \gets arg\max_{x' \in X}\{d_{Y}(h_{\\theta_t}(X_{t_b}),h_{\\theta_t}(x')) - \lambda_t d_{X}(X_{t_b},x')\}

    Proposed in `SenSeI: Sensitive Set Invariance for Enforcing Individual Fairness <https://arxiv.org/abs/2006.14168>`_

    Parameters
    --------------
        distance_x: inFairness.distances.Distance
            Distance metric in the input space
        distance_y: inFairness.distances.Distance
            Distance metric in the output space
        num_steps: int
            Number of update steps should the auditor perform to find worst-case examples
        lr: float
            Learning rate
    """

    def __init__(
        self, distance_x, distance_y, num_steps, lr, max_noise=0.1, min_noise=-0.1
    ):

        self.distance_x = distance_x
        self.distance_y = distance_y
        self.num_steps = num_steps
        self.lr = lr
        self.max_noise = max_noise
        self.min_noise = min_noise

        super().__init__()

    def generate_worst_case_examples(self, network, x, lambda_param, optimizer=None):
        """Generate worst case example given the input data sample batch `x`

        Parameters
        ------------
            network: torch.nn.Module
                PyTorch network model
            x: torch.Tensor
                Batch of input datapoints
            lambda_param: float
                Lambda weighting parameter as defined in the equation above
            optimizer: torch.optim.Optimizer, optional
                PyTorch Optimizer object. Default: torch.optim.Adam

        Returns
        ---------
            X_worst: torch.Tensor
                Worst case examples for the provided input datapoints
        """

        assert optimizer is None or issubclass(optimizer, torch.optim.Optimizer), (
            "`optimizer` object should either be None or be a PyTorch optimizer "
            + "and an instance of the `torch.optim.Optimizer` class"
        )

        freeze_network(network)
        lambda_param = lambda_param.detach()

        delta = Parameter(
            torch.rand_like(x) * (self.max_noise - self.min_noise) + self.min_noise
        )

        if optimizer is None:
            optimizer = torch.optim.Adam([delta], lr=self.lr)
        else:
            optimizer = optimizer([delta], lr=self.lr)

        for _ in range(self.num_steps):
            optimizer.zero_grad()
            x_worst = x + delta
            input_dist = self.distance_x(x, x_worst)
            out_x = network(x)
            out_x_worst = network(x_worst)

            out_dist = self.distance_y(out_x, out_x_worst)

            audit_loss = -(out_dist - lambda_param * input_dist)
            audit_loss.sum().backward()
            optimizer.step()

        unfreeze_network(network)

        return (x + delta).detach()

    def audit(
        self,
        network,
        X_audit,
        Y_audit,
        loss_fn,
        audit_threshold=None,
        lambda_param=None,
        confidence=0.95,
        optimizer=None,
    ):
        """Audit a model for individual fairness

        Parameters
        ------------
            network: torch.nn.Module
                PyTorch network model
            X_audit: torch.Tensor
                Auditing data samples. Shape: (B, *)
            Y_audit: torch.Tensor
                Auditing data samples. Shape: (B)
            loss_fn: torch.nn.Module
                Loss function
            audit_threshold: float, optional
                Auditing threshold to consider a model individually fair or not
                If `audit_threshold` is specified, the `audit` procedure determines
                if the model is individually fair or not.
                If `audit_threshold` is not specified, the `audit` procedure simply
                returns the mean and lower bound of loss ratio, leaving the determination
                of models' fairness to the user.
                Default=None
            lambda_param: float
                Lambda weighting parameter as defined in the equation above
            confidence: float, optional
                Confidence value. Default = 0.95
            optimizer: torch.optim.Optimizer, optional
                PyTorch Optimizer object. Default: torch.optim.SGD

        Returns
        ------------
            audit_response: inFairness.auditor.datainterface.AuditorResponse
                Audit response containing test statistics
        """

        assert optimizer is None or issubclass(optimizer, torch.optim.Optimizer), (
            "`optimizer` object should either be None or be a PyTorch optimizer "
            + "and an instance of the `torch.optim.Optimizer` class"
        )

        device = get_device(X_audit)

        if lambda_param is None:
            lambda_param = torch.tensor(1.0, device=device)
        elif isinstance(lambda_param, float):
            lambda_param = torch.tensor(lambda_param, device=device)

        if optimizer is None:
            optimizer = torch.optim.SGD

        X_worst = self.generate_worst_case_examples(
            network=network, x=X_audit, lambda_param=lambda_param, optimizer=optimizer
        )

        loss_ratio = self.compute_loss_ratio(
            X_audit=X_audit,
            X_worst=X_worst,
            Y_audit=Y_audit,
            network=network,
            loss_fn=loss_fn,
        )

        audit_response = self.compute_audit_result(
            loss_ratio, audit_threshold, confidence
        )

        return audit_response



import torch.nn


def freeze_network(network: torch.nn.Module):
    """Freeze network parameters.
    :param network: torch.nn.Module
    :type network: torch.nn.Module
    """
    for p in network.parameters():
        p.requires_grad = False


def unfreeze_network(network: torch.nn.Module):
    """Unfreeze network parameters.
    :param network: torch.nn.Module
    :type network: torch.nn.Module
    """
    for p in network.parameters():
        p.requires_grad = True



import torch

from typing import Tuple
from inFairness.postprocessing.data_ds import PostProcessingDataStore


class BasePostProcessing(object):
    """Base class for Post-Processing methods

    Parameters
    -------------
        distance_x: inFairness.distances.Distance
            Distance matrix in the input space
        is_output_probas: bool
            True if the `data_Y` (model output) are probabilities implying that
            this is a classification setting, and False if the `data_Y` are
            in euclidean space implying that this is a regression setting.
    """

    def __init__(self, distance_x, is_output_probas):

        self.distance_x = distance_x
        self.is_output_probas = is_output_probas
        self.datastore = PostProcessingDataStore(distance_x)

    @property
    def data(self):
        """Input and Output data used for post-processing

        Returns
        --------
            data: Tuple(torch.Tensor, torch.Tensor)
                A tuple of (X, Y) data points
        """
        return (self.datastore.data_X, self.datastore.data_Y)

    @property
    def distance_matrix(self):
        """Distance matrix

        Returns
        --------
            distance_matrix: torch.Tensor
                Matrix of distances of shape (N, N) where
                N is the number of data samples
        """
        return self.datastore.distance_matrix

    def add_datapoints(self, X: torch.Tensor, y: torch.Tensor):
        """Add datapoints to the post-processing method

        Parameters
        -----------
            X: torch.Tensor
                New input datapoints
            y: torch.Tensor
                New output datapoints
        """
        self.datastore.add_datapoints(X, y)

    def reset_datapoints(self):
        """Reset datapoints store back to its initial state"""
        self.datastore.reset()

    def postprocess(self, *args, **kwargs):
        raise NotImplementedError("postprocess method not implemented by class")

    def __get_yhat__(self):

        _, data_y = self.data

        if self.is_output_probas:
            y_hat = torch.log(data_y[:, :-1]) - torch.log(data_y[:, -1]).view(-1, 1)
            return y_hat
        else:
            return data_y



import numpy as np
import torch
from scipy.stats import logistic

from inFairness.utils import datautils
from inFairness.distances.mahalanobis_distance import MahalanobisDistances


class EXPLOREDistance(MahalanobisDistances):
    """Implements the Embedded Xenial Pairs Logistic Regression metric
    (EXPLORE) defined in Section 2.2 of Two Simple Ways to Learn Individual
    Fairness Metrics from Data.

    EXPLORE defines the distance in the input space to be of the form:

    .. math:: d_x(x_1, x_2) := \langle \phi(x_1) - \phi(x_2), \Sigma (\phi(x_1) - \phi(x_2)) \\rangle
    where :math:`\phi(x)` is an embedding map and :math:`\Sigma` is a semi-positive
    definite matrix.

    The metric expects the data to be in the form of triplets
    :math:`\{(x_{i_1}, x_{i_2}, y_i)\}_{i=1}^{n}` where :math:`y_i \in \{0, 1\}`
    indicates whether the human considers :math:`x_{i_1}` and :math:`x_{i_2}`
    comparable (:math:`y_i = 1` indicates comparable) or not.

    References
    -----------
        `Mukherjee, Debarghya, Mikhail Yurochkin, Moulinath Banerjee, and Yuekai Sun.
        "Two simple ways to learn individual fairness metrics from data." In
        International Conference on Machine Learning, pp. 7097-7107. PMLR, 2020.`
    """

    def __init__(self):
        super().__init__()

    def fit(self, X1, X2, Y, iters, batchsize, autoinfer_device=True):
        """Fit EXPLORE distance metric

        Parameters
        -----------
            X1: torch.Tensor
                first set of input samples
            X2: torch.Tensor
                second set of input samples
            Y: torch.Tensor
                :math:`y_i` vector containing 1 if corresponding elements from
                X1 and X2 are comparable, and 0 if not
            iters: int
                number of iterations of SGD to compute the :math:`\Sigma` matrix
            batchsize: int
                batch size of each iteration
            autoinfer_device: bool
                Should the distance metric be automatically moved to an appropriate
                device (CPU / GPU) or not? If set to True, it moves the metric
                to the same device `X1` is on. If set to False, keeps the metric
                on CPU.
        """

        X = datautils.convert_tensor_to_numpy(X1 - X2)
        Y = datautils.convert_tensor_to_numpy(Y)
        sigma = self.compute_sigma(X, Y, iters, batchsize)
        super().fit(sigma)

        if autoinfer_device:
            device = datautils.get_device(X1)
            super().to(device)

    def __grad_likelihood__(self, X, Y, sigma):
        """Computes the gradient of the likelihood function using sigmoidal link"""

        diag = np.einsum("ij,ij->i", np.matmul(X, sigma), X)
        diag = np.maximum(diag, 1e-10)
        prVec = logistic.cdf(diag)
        sclVec = 2.0 / (np.exp(diag) - 1)
        vec = (Y * prVec) - ((1 - Y) * prVec * sclVec)
        grad = np.matmul(X.T * vec, X) / X.shape[0]
        return grad

    def __projPSD__(self, sigma):
        """Computes the projection onto the PSD cone"""

        try:
            L = np.linalg.cholesky(sigma)
            sigma_hat = np.dot(L, L.T)
        except np.linalg.LinAlgError:
            d, V = np.linalg.eigh(sigma)
            sigma_hat = np.dot(
                V[:, d >= 1e-8], d[d >= 1e-8].reshape(-1, 1) * V[:, d >= 1e-8].T
            )
        return sigma_hat

    def compute_sigma(self, X, Y, iters, batchsize):

        N = X.shape[0]
        P = X.shape[1]

        sigma_t = np.random.normal(0, 1, P**2).reshape(P, P)
        sigma_t = np.matmul(sigma_t, sigma_t.T)
        sigma_t = sigma_t / np.linalg.norm(sigma_t)

        curriter = 0

        while curriter < iters:
            batch_idxs = np.random.choice(N, size=batchsize, replace=False)
            X_batch = X[batch_idxs]
            Y_batch = Y[batch_idxs]

            grad_t = self.__grad_likelihood__(X_batch, Y_batch, sigma_t)
            t = 1.0 / (1 + curriter // 100)
            sigma_t = self.__projPSD__(sigma_t - t * grad_t)

            curriter += 1

        sigma = torch.FloatTensor(sigma_t).detach()
        return sigma



import torch
import numpy as np

from abc import ABCMeta
from scipy.stats import norm
from inFairness.utils.datautils import convert_tensor_to_numpy
from inFairness.auditor.datainterface import AuditorResponse


class Auditor(metaclass=ABCMeta):
    """
    Abstract class for model auditors, e.g. Sensei or Sensr
    """

    def __init__(self):
        super(Auditor, self).__init__()

    def generate_worst_case_examples(self, *args, **kwargs):
        """Generates worst-case example for the input data sample batch"""
        raise NotImplementedError(
            "Method `generate_worst_case_examples` not implemented."
        )

    def compute_loss_ratio(self, X_audit, X_worst, Y_audit, network, loss_fn):
        """Compute the loss ratio of samples computed by solving gradient flow attack
        to original audit samples

        Parameters
        -------------
            X_audit: torch.Tensor
                Auditing samples. Shape (n_samples, n_features)
            Y_audit: torch.Tensor
                Labels of auditing samples. Shape: (n_samples)
            lambda_param: float
                Lambda weighting parameter as defined in the equation above

        Returns
        ---------
            loss_ratios: numpy.ndarray
                Ratio of loss for samples computed using gradient
                flow attack to original audit samples
        """

        with torch.no_grad():
            Y_pred_worst = network(X_worst)
            Y_pred_original = network(X_audit)

            loss_vals_adversarial = loss_fn(Y_pred_worst, Y_audit, reduction="none")
            loss_vals_original = loss_fn(Y_pred_original, Y_audit, reduction="none")

        loss_vals_adversarial = convert_tensor_to_numpy(loss_vals_adversarial)
        loss_vals_original = convert_tensor_to_numpy(loss_vals_original)

        loss_ratio = np.divide(loss_vals_adversarial, loss_vals_original)

        return loss_ratio

    def compute_audit_result(self, loss_ratios, threshold=None, confidence=0.95):
        """Computes auditing statistics given loss ratios and user-specified
        acceptance threshold

        Parameters
        -------------
            loss_ratios: numpy.ndarray
                List of loss ratios between worst-case and normal data samples
            threshold: float. optional
                User-specified acceptance threshold value
                If a value is not specified, the procedure simply returns the mean
                and lower bound of loss ratio, leaving the detemination of models'
                fairness to the user.
                If a value is specified, the procedure also determines if the model
                is individually fair or not.
            confidence: float, optional
                Confidence value. Default = 0.95

        Returns
        ----------
            audit_result: AuditorResponse
                Data interface with auditing results and statistics
        """

        loss_ratios = loss_ratios[np.isfinite(loss_ratios)]

        lossratio_mean = np.mean(loss_ratios)
        lossratio_std = np.std(loss_ratios)
        N = loss_ratios.shape[0]

        z = norm.ppf(confidence)
        lower_bound = lossratio_mean - z * lossratio_std / np.sqrt(N)

        if threshold is None:
            response = AuditorResponse(
                lossratio_mean=lossratio_mean,
                lossratio_std=lossratio_std,
                lower_bound=lower_bound,
            )
        else:
            tval = (lossratio_mean - threshold) / lossratio_std
            tval *= np.sqrt(N)

            pval = 1 - norm.cdf(tval)
            is_model_fair = False if pval < (1 - confidence) else True

            response = AuditorResponse(
                lossratio_mean=lossratio_mean,
                lossratio_std=lossratio_std,
                lower_bound=lower_bound,
                threshold=threshold,
                pval=pval,
                confidence=confidence,
                is_model_fair=is_model_fair,
            )

        return response

    def audit(self, *args, **kwargs):
        """Audit model for individual fairness"""
        raise NotImplementedError("Method not implemented")



import torch
import numpy as np
from functorch import vmap

from inFairness.distances.distance import Distance


class MahalanobisDistances(Distance):
    """Base class implementing the Generalized Mahalanobis Distances

    Mahalanobis distance between two points X1 and X2 is computed as:

    .. math:: \\text{dist}(X_1, X_2) = (X_1 - X_2) \\Sigma (X_1 - X_2)^{T}
    """

    def __init__(self):
        super().__init__()

        self.sigma = None
        self.device = torch.device("cpu")

    def to(self, device):
        """Moves distance metric to a particular device

        Parameters
        ------------
            device: torch.device
        """

        assert (
            self.sigma is not None
        ), "Please fit the metric before moving parameters to device"

        self.device = device
        self.sigma = self.sigma.to(self.device)

    def fit(self, sigma):
        """Fit Mahalanobis Distance metric

        Parameters
        ------------
            sigma: torch.Tensor
                    Covariance matrix
        """

        self.sigma = sigma

    @staticmethod
    def __compute_dist__(X1, X2, sigma):
        """Computes the distance between two data samples x1 and x2

        Parameters
        -----------
            X1: torch.Tensor
                Data sample of shape (n_features) or (N, n_features)
            X2: torch.Tensor
                Data sample of shape (n_features) or (N, n_features)

        Returns:
            dist: torch.Tensor
                Distance between points x1 and x2. Shape: (N)
        """

        # unsqueeze batch dimension if a vector is passed
        if len(X1.shape) == 1:
            X1 = X1.unsqueeze(0)
        if len(X2.shape) == 1:
            X2 = X2.unsqueeze(0)

        X_diff = X1 - X2
        dist = torch.sum((X_diff @ sigma) * X_diff, dim=-1, keepdim=True)
        return dist

    def forward(self, X1, X2, itemwise_dist=True):
        """Computes the distance between data samples X1 and X2

        Parameters
        -----------
            X1: torch.Tensor
                Data samples from batch 1 of shape (n_samples_1, n_features)
            X2: torch.Tensor
                Data samples from batch 2 of shape (n_samples_2, n_features)
            itemwise_dist: bool, default: True
                Compute the distance in an itemwise manner or pairwise manner.

                In the itemwise fashion (`itemwise_dist=False`), distance is
                computed between the ith data sample in X1 to the ith data sample
                in X2. Thus, the two data samples X1 and X2 should be of the same shape

                In the pairwise fashion (`itemwise_dist=False`), distance is
                computed between all the samples in X1 and all the samples in X2.
                In this case, the two data samples X1 and X2 can be of different shapes.

        Returns
        ----------
            dist: torch.Tensor
                Distance between samples of batch 1 and batch 2.

                If `itemwise_dist=True`, item-wise distance is returned of
                shape (n_samples, 1)

                If `itemwise_dist=False`, pair-wise distance is returned of
                shape (n_samples_1, n_samples_2)
        """

        if itemwise_dist:
            np.testing.assert_array_equal(
                X1.shape,
                X2.shape,
                err_msg="X1 and X2 should be of the same shape for itemwise distance computation",
            )
            dist = self.__compute_dist__(X1, X2, self.sigma)
        else:
            X1 = X1.unsqueeze(0) if len(X1.shape) == 2 else X1  # (B, N, D)
            X2 = X2.unsqueeze(0) if len(X2.shape) == 2 else X2  # (B, M, D)

            nsamples_x1 = X1.shape[1]
            nsamples_x2 = X2.shape[1]
            dist_shape = (-1, nsamples_x1, nsamples_x2)

            vdist = vmap(
                vmap(
                    vmap(self.__compute_dist__, in_dims=(None, 0, None)),
                    in_dims=(0, None, None),
                ),
                in_dims=(0, 0, None),
            )

            dist = vdist(X1, X2, self.sigma).view(dist_shape)

        return dist


class SquaredEuclideanDistance(MahalanobisDistances):
    """
    computes the squared euclidean distance as a special case of the mahalanobis distance where:

    .. math:: \\Sigma= I_{num_dims}
    """

    def __init__(self):
        super().__init__()
        self.num_dims_ = None

    def fit(self, num_dims: int):
        """Fit Square Euclidean Distance metric

        Parameters
        -----------------
            num_dims: int
                the number of dimensions of the space in which the Squared Euclidean distance will be used.
        """

        self.num_dims_ = num_dims
        sigma = torch.eye(self.num_dims_).detach()
        super().fit(sigma)



from inFairness.auditor.auditor import Auditor
from inFairness.auditor.sensei_auditor import SenSeIAuditor
from inFairness.auditor.sensr_auditor import SenSRAuditor

__all__ = [symb for symb in globals() if not symb.startswith("_")]



import torch

from dataclasses import dataclass


@dataclass
class AuditorResponse:
    """Class to store a result from the auditor"""

    lossratio_mean: float = None
    lossratio_std: float = None
    lower_bound: float = None
    threshold: float = None
    pval: float = None
    confidence: float = None
    is_model_fair: bool = None



import torch

from inFairness.postprocessing.distance_ds import DistanceStructure


class PostProcessingDataStore(object):
    """Data strucuture to hold the data used for post-processing

    Parameters
    -------------
        distance_x: inFairness.distances.Distance
            Distance metric in the input space
    """

    def __init__(self, distance_x):

        self.data_X = None
        self.data_Y = None
        self.n_samples = 0
        self.distance_ds = DistanceStructure(distance_x)

    @property
    def distance_matrix(self):
        """Distances between N data points. Shape: (N, N)"""
        return self.distance_ds.distance_matrix

    def add_datapoints_X(self, X: torch.Tensor):
        """Add datapoints to the input datapoints X

        Parameters
        ------------
            X: torch.Tensor
                New data points to add to the input data
                `X` should have the same dimensions as previous data
                along all dimensions except the first (batch) dimension
        """

        if self.data_X is None:
            self.data_X = X
        else:
            self.data_X = torch.cat([self.data_X, X], dim=0)

    def add_datapoints_Y(self, Y: torch.Tensor):
        """Add datapoints to the output datapoints Y

        Parameters
        ------------
            Y: torch.Tensor
                New data points to add to the output data
                `Y` should have the same dimensions as previous data
                along all dimensions except the first (batch) dimension
        """

        if self.data_Y is None:
            self.data_Y = Y
        else:
            self.data_Y = torch.cat([self.data_Y, Y], dim=0)

    def add_datapoints(self, X: torch.Tensor, Y: torch.Tensor):
        """Add new datapoints to the existing datapoints

        Parameters
        ------------
            X: torch.Tensor
                New data points to add to the input data
                `X` should have the same dimensions as previous data
                along all dimensions except the first (batch) dimension
            Y: torch.Tensor
                New data points to add to the output data
                `Y` should have the same dimensions as previous data
                along all dimensions except the first (batch) dimension
        """

        self.add_datapoints_X(X)
        self.add_datapoints_Y(Y)
        self.n_samples = self.n_samples + X.shape[0]
        self.distance_ds.build_distance_matrix(self.data_X)

    def reset(self):
        """Reset the data structure holding the data points for post-processing.
        Invoking this operation removes all datapoints and resets the state back
        to the initial state.
        """

        self.data_X = None
        self.data_Y = None
        self.n_samples = 0
        self.distance_ds.reset()



import torch


class DistanceStructure(object):
    """Data structure to store and track the distance matrix between data points

    Parameters
    -------------
        distance_x: inFairness.distances.Distance
            Distance metric in the input space
    """

    def __init__(self, distance_x):
        self.distance_x = distance_x
        self.distance_matrix = None

    def reset(self):
        """Reset the state of the data structure back to its initial state"""
        self.distance_matrix = None

    def build_distance_matrix(self, data_X):
        """Build the distance matrix between input data samples `data_X`

        Parameters
        -------------
            data_X: torch.Tensor
                Data points between which the distance matrix is to be computed
        """

        nsamples_old = (
            0 if self.distance_matrix is None else self.distance_matrix.shape[0]
        )
        nsamples_total = data_X.shape[0]
        device = data_X.device

        distance_matrix_new = torch.zeros(
            size=(nsamples_total, nsamples_total), device=device
        )

        if self.distance_matrix is not None:
            distance_matrix_new[:nsamples_old, :nsamples_old] = self.distance_matrix

        dist = (
            self.distance_x(
                data_X[nsamples_old:nsamples_total], data_X, itemwise_dist=False
            )
            .detach()
            .squeeze()
        )
        distance_matrix_new[nsamples_old:, :] = dist
        distance_matrix_new[:, nsamples_old:] = dist.T

        self.distance_matrix = distance_matrix_new.clone()



from typing import Iterable
import numpy as np
import torch
from sklearn.linear_model import LogisticRegression

from inFairness.distances import SensitiveSubspaceDistance
from inFairness.utils import datautils, validationutils


class LogisticRegSensitiveSubspace(SensitiveSubspaceDistance):
    """Implements the Softmax Regression model based fair metric as defined in Appendix B.1
    of "Training individually fair ML models with sensitive subspace robustness" paper.

    This metric assumes that the sensitive attributes are discrete and observed for a small subset
    of training data. Assuming data of the form :math:`(X_i, K_i, Y_i)` where :math:`K_i` is the
    sensitive attribute of the i-th subject, the model fits a softmax regression model to the data as:

    .. math:: \mathbb{P}(K_i = l\\mid X_i) = \\frac{\exp(a_l^TX_i+b_l)}{\\sum_{l=1}^k \\exp(a_l^TX_i+b_l)},\\ l=1,\\ldots,k

    Using the span of the matrix :math:`A=[a_1, \cdots, a_k]`, the fair metric is trained as:

    .. math:: d_x(x_1,x_2)^2 = (x_1 - x_2)^T(I - P_{\\text{ran}(A)})(x_1 - x_2)

    References
    -------------
        `Yurochkin, Mikhail, Amanda Bower, and Yuekai Sun. "Training individually fair
        ML models with sensitive subspace robustness." arXiv preprint arXiv:1907.00020 (2019).`
    """

    def __init__(self):
        super().__init__()
        self.basis_vectors_ = None

    def fit(
        self,
        data_X: torch.Tensor,
        data_SensitiveAttrs: torch.Tensor = None,
        protected_idxs: Iterable[int] = None,
        keep_protected_idxs: bool = True,
        autoinfer_device: bool = True,
    ):
        """Fit Logistic Regression Sensitive Subspace distance metric

        Parameters
        --------------
            data_X: torch.Tensor
                Input data corresponding to either :math:`X_i` or :math:`(X_i, K_i)` in the equation above.
                If the variable corresponds to :math:`X_i`, then the `y_train` parameter should be specified.
                If the variable corresponds to :math:`(X_i, K_i)` then the `protected_idxs` parameter
                should be specified to indicate the sensitive attributes.

            data_SensitiveAttrs: torch.Tensor
                Represents the sensitive attributes ( :math:`K_i` ) and is used when the `X_train` parameter
                represents :math:`X_i` from the equation above. **Note**: This parameter is mutually exclusive
                with the `protected_idxs` parameter. Specififying both the `data_SensitiveAttrs` and `protected_idxs`
                parameters will raise an error

            protected_idxs: Iterable[int]
                If the `X_train` parameter above represents :math:`(X_i, K_i)`, then this parameter is used
                to provide the indices of sensitive attributes in `X_train`. **Note**: This parameter is mutually exclusive
                with the `protected_idxs` parameter. Specififying both the `data_SensitiveAttrs` and `protected_idxs`
                parameters will raise an error

            keep_protected_indices: bool
                True, if while training the model, protected attributes will be part of the training data
                Set to False, if for training the model, protected attributes will be excluded
                Default = True

            autoinfer_device: bool
                Should the distance metric be automatically moved to an appropriate
                device (CPU / GPU) or not? If set to True, it moves the metric
                to the same device `X_train` is on. If set to False, keeps the metric
                on CPU.
        """

        if data_SensitiveAttrs is not None and protected_idxs is None:
            basis_vectors_ = self.compute_basis_vectors_data(
                X_train=data_X, y_train=data_SensitiveAttrs
            )

        elif data_SensitiveAttrs is None and protected_idxs is not None:
            basis_vectors_ = self.compute_basis_vectors_protected_idxs(
                data_X,
                protected_idxs=protected_idxs,
                keep_protected_idxs=keep_protected_idxs,
            )

        else:
            raise AssertionError(
                "Parameters `y_train` and `protected_idxs` are exclusive. Either of these two parameters should be None, and cannot be set to non-None values simultaneously."
            )

        super().fit(basis_vectors_)
        self.basis_vectors_ = basis_vectors_

        if autoinfer_device:
            device = datautils.get_device(data_X)
            super().to(device)

    def compute_basis_vectors_protected_idxs(
        self, data, protected_idxs, keep_protected_idxs=True
    ):

        dtype = data.dtype

        data = datautils.convert_tensor_to_numpy(data)
        basis_vectors_ = []
        num_attr = data.shape[1]

        # Get input data excluding the protected attributes
        protected_idxs = sorted(protected_idxs)
        free_idxs = [idx for idx in range(num_attr) if idx not in protected_idxs]
        X_train = data[:, free_idxs]
        Y_train = data[:, protected_idxs]

        self.__assert_sensitiveattrs_binary__(Y_train)

        coefs = np.array(
            [
                LogisticRegression(solver="liblinear", penalty="l1")
                .fit(X_train, Y_train[:, idx])
                .coef_.squeeze()
                for idx in range(len(protected_idxs))
            ]
        )  # ( |protected_idxs|, |free_idxs| )

        if keep_protected_idxs:
            # To keep protected indices, we add two basis vectors
            # First, with logistic regression coefficients with 0 in
            # protected indices. Second, with one-hot vectors with 1 in
            # protected indices.

            basis_vectors_ = np.empty(shape=(2 * len(protected_idxs), num_attr))

            for i, protected_idx in enumerate(protected_idxs):

                protected_basis_vector = np.zeros(shape=(num_attr))
                protected_basis_vector[protected_idx] = 1.0

                unprotected_basis_vector = np.zeros(shape=(num_attr))
                np.put_along_axis(
                    unprotected_basis_vector, np.array(free_idxs), coefs[i], axis=0
                )

                basis_vectors_[2 * i] = unprotected_basis_vector
                basis_vectors_[2 * i + 1] = protected_basis_vector
        else:
            # Protected indices are to be discarded. Therefore, we can
            # simply return back the logistic regression coefficients
            basis_vectors_ = coefs

        basis_vectors_ = torch.tensor(basis_vectors_, dtype=dtype).T
        basis_vectors_ = basis_vectors_.detach()

        return basis_vectors_

    def compute_basis_vectors_data(self, X_train, y_train):

        dtype = X_train.dtype

        X_train = datautils.convert_tensor_to_numpy(X_train)
        y_train = datautils.convert_tensor_to_numpy(y_train)

        self.__assert_sensitiveattrs_binary__(y_train)

        basis_vectors_ = []
        outdim = y_train.shape[-1]

        basis_vectors_ = np.array(
            [
                LogisticRegression(solver="liblinear", penalty="l1")
                .fit(X_train, y_train[:, idx])
                .coef_.squeeze()
                for idx in range(outdim)
            ]
        )

        basis_vectors_ = torch.tensor(basis_vectors_, dtype=dtype).T
        basis_vectors_ = basis_vectors_.detach()

        return basis_vectors_

    def __assert_sensitiveattrs_binary__(self, sensitive_attrs):

        assert validationutils.is_tensor_binary(
            sensitive_attrs
        ), "Sensitive attributes are required to be binary to learn the metric. Please binarize these attributes before fitting the metric."



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="inFairness",
    packages=[
        "inFairness",
        *["inFairness." + p for p in find_packages(where="./inFairness")],
    ],
    package_dir={"": ".",},
    install_requires=[
        "torch>=1.11.0",
        "numpy>=1.21.6",
        "scikit-learn>=0.24.2",
        "pandas>=1.3.5",
        "scipy>=1.5.4",
        "functorch~=0.1.1"
    ],
    description="inFairness is a Python package to train and audit individually fair PyTorch models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="0.2.1",
    url="https://github.com/IBM/inFairness",
    author="IBM Research",
    author_email="mayank.agarwal@ibm.com, aldo.pareja@ibm.com, onkarbhardwaj@ibm.com, mikhail.yurochkin@ibm.com",
    keywords=[
        "individual fairness",
        "ai fairness",
        "trustworthy ai",
        "machine learning",
    ],
    python_requires=">=3.8",
)
